from .file_tools import read_yaml
import os

GLOBALS = {"INIT": {},
           "CONFIGS": {},
           "DB": {},
           "API": {},
           "MONGO": "mongo",
           "POSTGRES": "postgres"}

## Directly get/set/delete/load the GLOBALS

def get_globals(key= ""):
    if key:
        return GLOBALS.get(key, "")
    else:
        return GLOBALS

def set_globals(key: str, value):
    GLOBALS[key] = value

def delete_globals(key: str):
    del GLOBALS[key]

def load_globals_from_yaml(key: str, config_fn="", fn="", raise_exception=False):
    """
    Load the GLOBALS from the yaml file.
    
    Args:
        key (str): The key of the GLOBALS.
        config_fn (str): The config file name.
        fn (str): The file name.
        raise_exception (bool): Whether to raise an exception when the indicator is False.
        
    Returns:
        res (dict): The result.
    """
    if config_fn:
        res = read_yaml(f"{os.getcwd()}/configs/{config_fn}")
    else:
        res = read_yaml(fn)
        
    if res["indicator"]:
        set_globals(key, res["content"])
    elif raise_exception:
        raise Exception(res["message"])

    return res

## =========================================================

## Directly get/set/delete the CONFIGS

def get_configs(key= ""):
    if key:
        return GLOBALS["CONFIGS"].get(key, "")
    else:
        return GLOBALS["CONFIGS"]

def set_configs(key: str, value):
    GLOBALS["CONFIGS"][key] = value

def delete_configs(key: str):
    del GLOBALS["CONFIGS"][key]

## =========================================================

## Directly get/set/delete the DB

def get_db(db_type="", key= ""):
    if db_type and key:
        return GLOBALS["DB"][db_type].get(key, None)
    elif db_type:
        return GLOBALS["DB"].get(db_type, {})
    else:
        return GLOBALS["DB"]

def set_db(db_type, key: str, value):
    if db_type not in GLOBALS["DB"]:
        GLOBALS["DB"][db_type] = {}
    GLOBALS["DB"][db_type][key] = value

def delete_db(db_type, key: str):
    del GLOBALS["DB"][db_type][key]


import traceback
import logging
from functools import wraps
from flask import request, g, Blueprint
from .inspector_tools import api_use_inspector

def handler(endpoint: str, blueprint: Blueprint, **options):
    """
    This decorator is used to handle the request from the client.
    Parameters:
        endpoint: The endpoint of the request.
        blueprint: The blueprint of the request.
        **options: The options of the request.
    
    Returns:
        func: The function to handle the request with the following parameters:
        - input_request: The input request from the client.
        - DB: The database objects (Postgres Handler).
        - CONFIGS: The configurations of the server.
        - API_CONFIGS: The configurations of the API.
        - USER_DATA: The information of the user.
    """
    def decorator(func):
        @blueprint.route(endpoint, **options)
        @wraps(func)
        def wrapper(**kwargs):
            if api_use_inspector(request):
                user_list = g.get("user_list", [])
                if len(user_list) == 1:
                    user_data = user_list[0]
                else:
                    user_data = {}
                params = {"input_request": g.get("input_request", {}),
                          "API_CONFIGS":   g.get("API_CONFIGS", {}),
                          "CONFIGS":       g.get("CONFIGS", {}),
                          "DB":            g.get("DB", {}),
                          "USER_DATA":     user_data}
                params.update(kwargs)
                return func(**params)
            else:
                raise DevelopmentError("The server_api.yaml is not used, so 'handler' decorator is invalid.")
        return wrapper
    return decorator

def TrueResponse(message: str, content=None) -> dict:
    """
    This function is used to return the message to the client with {"indicator": True, "message": message, "content": content}
    
    Args:
        message (str): The message to return to the client.
        content (any): The content to return to the client.
    
    Returns:
        dict: The message to return to the client.
    """
    return {"indicator": True, "message": message, "content": content}

def FalseResponse(message: str, content=None) -> dict:
    """
    This function is used to return the message to the client with {"indicator": False, "message": message}
    
    Args:
        message (str): The message to return to the client.
    
    Returns:
        dict: The message to return to the client.
    """
    raise IndicatorFalseException(message, content=content)

class IndicatorFalseException(Exception):
    """
    This class is used to raise errors in the server.
    If developers want to return indicator False, they should raise this error.
    Then, server will return the message to the client with {"indicator": False, "message": message}
    """
    def __init__(self, message, content=None):
        super().__init__()
        self.message = message
        self.content = content
        self.filename, self.lineno, self.function, self.text = traceback.extract_stack()[-2]
    
    def __str__(self):
        return self.message
    
    def get_json_response(self):
        return {"indicator": False, "message": self.message, "content": self.content}

class DevelopmentError(Exception):
    def __init__(self, message):
        super().__init__()
        self.message = message
        self.filename, self.lineno, self.function, self.text = traceback.extract_stack()[-2]
    
    def __str__(self):
        return self.message

class EmptyInputRequest:
    """
    This class is used to handle the empty input request.
    """
    def __init__(self, key: str) -> None:
        self.key = key
    
    def __str__(self) -> str:
        return f"(EmptyInputRequest) {self.key}"
    
    def __repr__(self) -> str:
        return self.__str__()

def error_handler(e, loggerIndicatorFalse=False):
    """
    This function is used to handle the error in the server.

    Args:
        e (Exception): The exception to handle.
        loggerIndicatorFalse (bool): If True, the server will log the error message when the class is IndicatorFalseException.
    """
    if isinstance(e, IndicatorFalseException):
        if loggerIndicatorFalse:
            logging.warning(f"\n============ Server API Indicator False ============")
            logging.warning(traceback.format_exc())
            logging.warning("=====================================================\n")
        return e.get_json_response()
    else:
        logging.warning(f"\n====================================================")
        logging.warning(traceback.format_exc())
        logging.warning("=====================================================\n")
        return {"indicator": False, "message": str(e), "content": None}

def api_route_register(app, blueprint: Blueprint):
    """
    This function is used to register the blueprint to the app.
    The url_prefix is f"/api/{blueprint.name}".
    """
    app.register_blueprint(blueprint, url_prefix=f"/api/{blueprint.name}")



import bcrypt

# =============== Password Handler ===============

def hash_password(raw_pw: str) -> str:
    salt = bcrypt.gensalt(rounds=12)
    hash_value = bcrypt.hashpw(raw_pw.encode(), salt)
    return hash_value.decode()

def check_password(raw_pw: str, hashed_pw: str) -> bool:
    return bcrypt.checkpw(raw_pw.encode(), hashed_pw.encode())


import yaml
import configparser

def read_yaml(fn: str) -> dict:
	"""
	This is the helper function that reads the config yaml
	file and returns a config dictionary
	- Input:
		fn: filename
	- Returns:
		dict: a config dictionary
	"""
	try:
		with open(fn, 'r', encoding="utf-8") as file:
			# The FullLoader parameter handles the conversion from YAML
			# scalar values to Python the dictionary format
			config_yaml = yaml.load(file, Loader=yaml.FullLoader)
			return {"indicator": True, "message": "Load a yaml file successfully", "content":config_yaml}
	except Exception as e:
		return {"indicator": False, "message": str(e)}

def write_yaml(fn: str, content: dict, sort_keys=True) -> dict:
	"""
	This is the helper function that writes to the config yaml and 
	returns the status indicating whether the operation is successful or not.
	- Input:
		* fn: filename
		* content: resulting yaml to be added 
	- Returns:
		A dictionary containing the following keys
		* indicator: True, False whether the operation is completed
		* message: message of operation
	"""
	indicator = True
	message = "Write to yaml successful"
	try:
		with open(fn, "w") as file:
			yaml.dump(content, file, allow_unicode = True, sort_keys=sort_keys)
	except Exception as e:
		indicator = False
		message = str(e) 
	
	return {"indicator":indicator, "message":message}

def read_ini(fn: str) -> dict:
	"""
	This is the helper function that reads the config ini
	Parameters:
		fn: filename
	Returns:
		A dictionary containing the following keys
		* indicator: True, False whether the operation is completed
		* message: message of operation
	"""
	try:
		config_ini = configparser.ConfigParser()
		config_ini.read(fn)
		return {"indicator": True, "message": "Load an ini file successfully", "content": config_ini}
	except Exception as e:
		return {"indicator": False, "message": str(e)}


from abc import ABC, abstractmethod
from flask import g, jsonify, Request, Response
from flask_jwt_extended import get_jwt, verify_jwt_in_request
import json, logging
from .api_tools import FalseResponse, DevelopmentError, EmptyInputRequest
from .global_tools import get_globals
from .inspector_tools import global_use_inspector, api_use_inspector, jwt_use_inspector, db_use_inspector

## =============== Original Middleware ===============

class Middleware(ABC):
    """
    Middleware is an abstract class that defines each process of middleware.
    
    The methods "process_request" and "process_response" must be implemented.
    """
    @abstractmethod
    def process_request(self, request: Request, stop):
        """
        Process a request.

        Args:
            request (Request): The request to process.
            stop (function): A function that stops the request from continuing through the middleware list.
        """
        pass

    @abstractmethod
    def process_response(self, response: Response, stop):
        """
        Process a response.

        Args:
            response (Response): The response to process.
            stop (function): A function that stops the response from continuing through the middleware list.
        """
        pass

class MiddlewareSystem:
    """
    MiddlewareSystem is a class that allows you to add middleware to your application.
    """
    def __init__(self):
        self.middleware_list = [_PassGlobal(), _InputRequest(), _JWTMiddleware()]
        self.continue_request = True
        self.continue_response = True

    def use(self, middleware: Middleware):
        """
        Add a middleware to the middleware list.
        
        Args:
            middleware (Middleware): The middleware to add to the list.
        """
        self.middleware_list.append(middleware)
        return self

    def process_request(self, request: Request):
        """
        Process a request through the middleware list.

        Args:
            request (Request): The request to process.
        """
        for middleware in self.middleware_list:
            if not self.continue_request:
                break
            middleware.process_request(request, lambda: self.stop_request())
        self.continue_request = True

    def process_response(self, response: Response):
        """
        Process a response through the middleware list.

        Args:
            response (Response): The response to process.
        """
        for middleware in self.middleware_list:
            if not self.continue_response:
                break
            middleware.process_response(response, lambda: self.stop_response())
        self.continue_response = True

    def stop_request(self):
        """
        Stop the request from continuing through the middleware list.
        """
        self.continue_request = False
    
    def stop_response(self):
        """
        Stop the response from continuing through the middleware list.
        """
        self.continue_response = False

## =============== Private Middleware ===============

class _PassGlobal(Middleware):
    def process_request(self, request, stop):
        if global_use_inspector():
            GLOBALS = get_globals()
            for key, value in GLOBALS.items():
                setattr(g, key, value)
        
            if api_use_inspector(request):
                if request.method in ["GET", "POST", "PATCH", "PUT", "DELETE"]:
                    g.API_CONFIGS = self.get_api_configs(request)
                else:
                    stop()
            else:
                stop()
        else:
            logging.warning("The GLOBALS is not loaded.")
            stop()

    def process_response(self, response, stop):
        pass

    def get_api_configs(self, request):
        api_configs = g.get("API", {})
        route_list = request.path.strip("/").split("/")
        service = route_list[1]
        operation = route_list[2]
        return api_configs[service][operation][request.method]

class _InputRequest(Middleware):
    def process_request(self, request, stop):
        try:
            if request.method in ["GET", "DELETE"]:
                input_request = request.args.to_dict()
            elif request.method in ["POST", "PATCH", "PUT"]:
                input_request = request.get_json()
            elif request.method in ["OPTIONS"]:
                input_request = {}
                stop()
            else:
                input_request = {}
                stop()
        except:
            input_request = {}
        g.input_request = input_request
        
    def process_response(self, response, stop):
        pass

class _JWTMiddleware(Middleware):
    def process_request(self, request, stop):
        if jwt_use_inspector():
            api_configs = g.get("API_CONFIGS", {})
            if api_configs.get("Authentication", False):
                verify_jwt_in_request()

    def process_response(self, response, stop):
        pass

## =============== Auth Middleware ===============

class _AuthMiddleware(Middleware):
    """
    AuthMiddleware is a middleware about authentication.
    """
    @classmethod
    def get_user_list(cls, DH_OBJ: object, target: str, unique_key: str) -> dict:
        """
        Get the user data from database.
        
        Args:
            DH_OBJ (object): The database handler ( Instantiated Object ).
            target (str): The table or collection name.
            unique_key (str): The unique key of the table or collection.
        
        Returns:
            dict: The user data or the exception.
        """
        user_list = g.get("user_list", None)
        if not user_list:
            user_list = cls._query_data(DH_OBJ, target, unique_key)
            g.user_list = user_list
        return user_list
    
    @staticmethod
    def _query_data(DH_OBJ: object, target: str, unique_key: str) -> dict:
        """
        Query the user data from database.
        
        Args:
            DH_OBJ (object): The database handler ( Instantiated Object ).
            target (str): The table or collection name.
            unique_key (str): The unique key of the table or collection.
        
        Returns:
            dict: The user data or the exception.
        """
        token_data: dict = get_jwt()
        uid = token_data.get(unique_key, "")
        if not uid:
            raise DevelopmentError(f"Token doesn't have the value from {unique_key}.")
        
        if DH_OBJ.db_type == get_globals("POSTGRES"):
            condition = [(f"{unique_key}=", token_data.get(unique_key, ""))]
            query_res = DH_OBJ.get_data(table=target, conditional_rule_list=condition)
        elif DH_OBJ.db_type == get_globals("MONGO"):
            condition = {unique_key: token_data.get(unique_key, "")}
            query_res = DH_OBJ.get_data(schema_name=target, conditions=condition)
        else :
            raise DevelopmentError("DB type is not supported.")
        
        if not query_res["indicator"]:
            return FalseResponse(query_res["message"])

        for item in query_res["formatted_data"]:
            del item["password"]
        return query_res["formatted_data"]

class UserExistence(_AuthMiddleware):
    """
    This middleware is used to check if the user exists in database.

    Args:
        DH_OBJ (object): The database handler ( Instantiated Object ).
        target (str): The table or collection name.
        unique_key (str): The unique key of the table or collection.
    """
    def __init__(self, DH_OBJ, target: str, unique_key: str):
        self.DH_OBJ = DH_OBJ
        self.target = target
        self.unique_key = unique_key
    
    def process_request(self, request, stop):
        if jwt_use_inspector() and api_use_inspector(request) and db_use_inspector():
            api_configs = g.get("API_CONFIGS", {})
            if api_configs.get("Authentication", False):
                user_list = UserExistence.get_user_list(self.DH_OBJ, self.target, self.unique_key)
            
                if len(user_list) == 0:
                    token_data:dict = get_jwt()
                    return FalseResponse(f"This account ({token_data[self.unique_key]}) is not in database.")
                elif len(user_list) > 1:
                    return FalseResponse("User ID Query Error: Length > 1")
    
    def process_response(self, response, stop):
        pass

class UserDatabasePermission(_AuthMiddleware):
    """
    This middleware is used to check if the user has the same permission as the permission in database.

    Args:
        DH_OBJ (object): The database handler ( Instantiated Object ).
        target (str): The table or collection name.
        unique_key (str): The unique key of the table or collection.
    """
    def __init__(self, DH_OBJ, target: str, unique_key: str):
        self.DH_OBJ = DH_OBJ
        self.target = target
        self.unique_key = unique_key
    
    def process_request(self, request, stop):
        if jwt_use_inspector() and api_use_inspector(request) and db_use_inspector():
            api_configs = g.get("API_CONFIGS", {})
            if api_configs.get("Authentication", False):
                user_list = UserDatabasePermission.get_user_list(self.DH_OBJ, self.target, self.unique_key)
                if len(user_list) != 1:
                    return FalseResponse("User ID Query Error: Length != 1")
                
                token_permission = get_jwt().get("permission", "")
                db_permission = user_list[0].get("permission", "")
                if token_permission != db_permission:
                    return FalseResponse("Your token permission is not consistent with permission in database.")

    def process_response(self, response, stop):
        pass

class UserApiPermission(_AuthMiddleware):
    """
    This middleware is used to check if the user has the permission to call this API.
    """
    def process_request(self, request, stop):
        if jwt_use_inspector() and api_use_inspector(request):
            api_configs = g.get("API_CONFIGS", {})
            if api_configs.get("Authentication", False):
                token_permission = get_jwt().get("permission", "")
                if token_permission not in api_configs["Permission"]:
                    return FalseResponse("Your don't have the permission to call this API.")

    def process_response(self, response, stop):
        pass

class UserStatusCheck(_AuthMiddleware):
    """
    This middleware is used to check the account status.
    
    Args:
        DH_OBJ (object): The database handler object.
        target (str): The table or collection name.
        unique_key (str): The unique key of the table or collection.
        status_key (str): The status key. Default: "status".
        excluding_status (list): The excluding status.
    """
    def __init__(self, DH_OBJ, target: str, unique_key: str, status_key="status", excluding_status=[]):
        self.DH_OBJ = DH_OBJ
        self.target = target
        self.unique_key = unique_key
        self.status_key = status_key
        self.excluding_status = excluding_status
    
    def process_request(self, request: Request, stop):
        if jwt_use_inspector() and api_use_inspector(request) and db_use_inspector():
            api_configs = g.get("API_CONFIGS", {})
            if api_configs.get("Authentication", False):
                user_list = UserStatusCheck.get_user_list(self.DH_OBJ, self.target, self.unique_key)
                if len(user_list) != 1:
                    return FalseResponse("User ID Query Error: Length != 1")
                
                status = user_list[0][self.status_key]
                if status in self.excluding_status:
                    return FalseResponse(f"This user is {status.capitalize()}")
    
    def process_response(self, response: Response, stop):
        pass

## =============== SafeKeys Middleware ===============

class _SafeKeysMiddleware(Middleware):
    @classmethod
    def get_input_request(cls, request):
        input_request = g.get("input_request", {})
        api_configs = g.get("API_CONFIGS", {})
        constructed_safe_keys = g.get("constructed_safe_keys", False)
        valid_methods = ["GET", "POST", "PATCH", "PUT", "DELETE"]

        if (request.method in valid_methods) and (not constructed_safe_keys) :
            input_request = cls._construct_input_request(input_request, api_configs)
            g.input_request = input_request
            g.constructed_safe_keys = True
        return input_request

    @classmethod
    def _construct_input_request(cls, input_request, api_configs):
        input_key_list = input_request.keys()
        opt_key_list = api_configs["Optionals"]
        opt_val_data = api_configs["OptionalDefaults"]
        if type(opt_val_data) == list:
            cls.check_api_configs_validity(api_configs)
            update_dict = {}
            for index in range(len(opt_key_list)):
                if opt_key_list[index] not in input_key_list:
                    update_dict[opt_key_list[index]] = opt_val_data[index]
            input_request.update(update_dict)
        elif type(opt_val_data) == dict:
            if not set(opt_val_data.keys()).issubset(set(opt_key_list)):
                raise DevelopmentError("The keys in OptionalDefaults should be included in Optionals.")
            for key in opt_key_list:
                if key not in input_key_list:
                    if key in opt_val_data:
                        input_request[key] = opt_val_data[key]
                    else:
                        input_request[key] = EmptyInputRequest(key)
        return input_request

    @staticmethod
    def check_api_configs_validity(api_configs):
        if not len(api_configs["Optionals"]) == len(api_configs["OptionalDefaults"]):
            raise DevelopmentError("The length of 'Optionals' and 'OptionalDefaults' in api config file should be the same.")

class SafeMandatoryKeys(_SafeKeysMiddleware):
    """
    This middleware is used to check if the input request has the mandatory keys.
    """
    def process_request(self, request, stop):
        if api_use_inspector(request):
            input_request = SafeMandatoryKeys.get_input_request(request)
            input_keys_list = input_request.keys()
            api_configs = g.get("API_CONFIGS", {})
            lack_list = []
            for item in api_configs["Mandatory"]:
                if item not in input_keys_list:
                    lack_list.append(item)
            if len(lack_list) > 0:
                return FalseResponse(f"No input: {','.join(lack_list)}")

    def process_response(self, response, stop):
        pass

class SafeInputType(_SafeKeysMiddleware):
    """
    This middleware is used to check if the input type is correct.
    """
    def process_request(self, request, stop):
        if api_use_inspector(request):
            input_request = SafeInputType.get_input_request(request)
            api_configs = g.get("API_CONFIGS", {})
            use_inspector = api_configs.get("InputInspect", False)
            input_type_configs = api_configs.get("InputType", None)
            if use_inspector:
                if input_type_configs == None:
                    raise DevelopmentError("You should set the InputType in server_api.yaml if you want to use InputInspect.")
                
                for inspect_key, inspect_type_info_str in input_type_configs.items():
                    inspect_type_info = inspect_type_info_str.split("::")
                    nullable = False
                    if len(inspect_type_info) > 1:
                        inspect_type = inspect_type_info[0]
                        nullable = "nullable" in inspect_type_info[1:]
                    else:
                        inspect_type = inspect_type_info[0]
                    
                    try:
                        inspect_value = input_request[inspect_key]
                    except KeyError as e:
                        raise DevelopmentError(f"The InputType of server_api.yaml not have the key '{inspect_key}'")
                    
                    if inspect_value == None:
                        if not nullable:
                            return FalseResponse(f"The input type of '{inspect_key}' should not be None but *{inspect_type}")
                    elif not isinstance(inspect_value, EmptyInputRequest):
                        if not self.input_type_inspector(inspect_type, inspect_value):
                            return FalseResponse(f"The input type of '{inspect_key}' should not be *{type(inspect_value).__name__} but *{inspect_type}")

    def input_type_inspector(self, inspect_type: str, inspect_value: str) -> bool:
        flag = True
        if inspect_type in ["json", "json-list", "json-dict"]:
                try:
                    loads = json.loads(inspect_value)
                    if inspect_type == "json-list":
                        flag = (type(loads) == list)
                    elif inspect_type == "json-dict":
                        flag = (type(loads) == dict)
                except:
                    flag = False
        else:
            flag = ( eval(inspect_type) == type(inspect_value) )
        return flag

    def process_response(self, response, stop):
        pass

# =============== Response Middleware ===============

class JsonifyResponse(Middleware):
    """
    This middleware is used to jsonify the response.
    """
    def process_request(self, request, stop):
        pass

    def process_response(self, response, stop):
        if isinstance(response.data, dict):
            response.data = jsonify(response.data)
        return response



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), "r") as f:
    requirements = f.read().split("\n")

setuptools.setup(
    name="feliz",
    version="0.0.8",
    author="Vincent Wu, Linga Chen, Brian Yin",
    author_email="zenith3092@gmail.com",
    description="A framework designed to assist in using Flask and conveniently developing RESTful APIs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zenith3092/feliz",
    license="MIT",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    project_urls={
        "Source": "https://github.com/zenith3092/feliz",
    },
)


import logging
import datetime
import time
import os
import json
import importlib
from abc import ABC, abstractmethod
from flask import Flask
from flask.json.provider import DefaultJSONProvider, _default
from flask_jwt_extended import JWTManager
from flask_cors import CORS
from .file_tools import read_ini
from .global_tools import load_globals_from_yaml, get_configs, set_db, get_globals, set_configs
from .inspector_tools import jwt_use_inspector, cors_use_inspector, db_use_inspector
from .api_tools import api_route_register

## =============== Original Initialware =============== ##

class Initialware(ABC):
    """
    Initialware is an abstract class that defines each process of initialization.
    
    The method "process" must be implemented.
    """
    @abstractmethod
    def process(self, prev_data: dict):
        """
        Process the initialization.
        
        Args:
            prev_data (dict): The previous data.
        """
        pass
    
    def next(self, prev_data):
        """
        Call the process method.
        """
        return self.process(prev_data)

class InitialwareSystem:
    """
    The InitialwareSystem class allows you to execute a series of initialwares.
    """
    def __init__(self):
        self.initialwares = []

    def use(self, initialware: Initialware):
        """
        Add an initialware to the initialware list.
        
        Args:
            initialware (Initialware): The initialware to add to the list.
        
        Returns:
            InitialwareSystem: The InitialwareSystem instance.
        """
        self.initialwares.append(initialware)
        return self

    def execute(self, app: Flask, first_data: dict = {}):
        """
        Execute the initialware list.

        Args:
            first_data (dict): The first input data. (The key "app" is reserved.)
        """
        data = first_data
        if "app" in data:
            raise Exception("The key 'app' is reserved in InitialwareSystem.")
        data["app"] = app
        start = time.time()
        for initialware in self.initialwares:
            data = initialware.next(data)
        end = time.time()
        logging.info(f" InitialwareSystem Execution Time: {end - start} seconds")

        server_config = get_configs("SERVER")
        logging.info(f" ***** {server_config['NAME']} is running on {server_config['HOST']}:{server_config['PORT']} ***** ")

## =============== Initialware Tools =============== ##

class ImportGlobals(Initialware):
    """
    Import the GLOBALS.
    """
    def __init__(self, config_fn="private/server_config.yaml"):
        self.config_fn = config_fn

    def process(self, prev_data):
        """
        Import the GLOBALS.
        """
        load_config_res = load_globals_from_yaml(key="CONFIGS", config_fn=self.config_fn)
        if load_config_res["indicator"]:
            config_data = load_config_res["content"]
            if ("API" in config_data) and (config_data["API"].get("API_ENABLE", False)):
                load_api_res = load_globals_from_yaml(key="API", config_fn=config_data["API"].get("API_FILE", ""))
                if not load_api_res["indicator"]:
                    logging.warning(f'Load API Config File Error: {load_api_res["message"]}')
        else:
            logging.warning(f'Load Config File Error: {load_config_res["message"]}')
        return prev_data

class JwtInitialware(Initialware):
    """
    The JwtInitialware class is used to initialize the JWT.
    """
    def process(self, prev_data):
        """
        Initialize the JWT.
        """
        if jwt_use_inspector():
            JWT_CONFIGS = get_configs("JWT")
            EXPIRE_HOURS = JWT_CONFIGS["EXPIRE_HOURS"]
            if EXPIRE_HOURS == "INFINITE":
                JWT_CONFIGS.update({"JWT_ACCESS_TOKEN_EXPIRES": False})
            else:
                JWT_CONFIGS.update({"JWT_ACCESS_TOKEN_EXPIRES": datetime.timedelta(hours=JWT_CONFIGS["EXPIRE_HOURS"])})
            prev_data["app"].config.update(JWT_CONFIGS)
            jwt = JWTManager(prev_data["app"])
            
            RETURN_MESSAGE = JWT_CONFIGS.get("MESSAGE", {})

            @jwt.unauthorized_loader
            def unauthorized_callback(error):
                return {"indicator": False, "message": RETURN_MESSAGE.get("UNAUTHORIZED", "Missing JWT token")}

            @jwt.invalid_token_loader
            def invalid_token_callback(error):
                return {"indicator": False, "message": RETURN_MESSAGE.get("INVALID_TOKEN", "Invalid JWT token")}

            @jwt.revoked_token_loader
            def revoked_token_callback(error):
                return {"indicator": False, "message": RETURN_MESSAGE.get("REVOKED_TOKEN", "Revoked JWT token")}

            @jwt.expired_token_loader
            def expired_token_callback(error, expired_token):
                return {"indicator": False, "message": RETURN_MESSAGE.get("EXPIRED_TOKEN", "Expired JWT token")}
        return prev_data

class CorsInitialware(Initialware):
    """
    The CorsInitialware class is used to initialize the CORS.
    Please refer to the Flask-CORS documentation for more information.
    https://flask-cors.readthedocs.io/en/latest/api.html
    
    Args:
        settings (dict)
    """
    def __init__(self, settings={}):
        self.kwargs = settings
    
    def process(self, prev_data):
        """
        Initialize the CORS.
        """
        if cors_use_inspector():
            CORS(prev_data["app"], **self.kwargs)
        return prev_data

class _DatabaseInitialware(Initialware):
    """
    The DatabaseInitialware class is used to initialize the database.
    """
    MONGO = get_globals("MONGO")
    POSTGRES = get_globals("POSTGRES")
    INI_CONFIG = {}
    
    @classmethod
    def get_ini_configs(cls):
        """
        Get the ini configs.
        """
        indicator = True
        message = "Success"

        if len(cls.INI_CONFIG.keys()) == 0:
            cls.INI_CONFIG = {cls.MONGO: {}, cls.POSTGRES: {}}

            DB_CONFIGS = get_configs("DB")
            ini_res = read_ini(f"{os.getcwd()}/configs/{DB_CONFIGS['INI_FILE']}")
            if ini_res["indicator"]:
                ini_data = ini_res["content"]
                for section in ini_res["content"].sections():
                    if ini_data[section]["db_type"] in cls.INI_CONFIG.keys():
                        cls.INI_CONFIG[ini_data[section]["db_type"]][section] = ini_data[section]
                    else:
                        indicator = False
                        message = f'({section}) {ini_data[section]["db_type"]} is not a valid db_type.'
            else:
                indicator = False
                message = f'{ini_res["message"]}'
        return {"indicator": indicator, "message": message, "content": cls.INI_CONFIG}

class MongoInitialware(_DatabaseInitialware):
    """
    The MongoInitialware class is used to initialize the mongo database.

    Args:
        mongo_handler_class (class): The mongo handler class.
        mongo_models (dict): {alias: models}
    """
    def __init__(self, mongo_handler_class, mongo_models):
        self.mongo_handler_class = mongo_handler_class
        self.mongo_models = mongo_models
    
    def process(self, prev_data):
        """
        Initialize the mongo database.
        """
        if db_use_inspector():
            get_ini_res = self.get_ini_configs()
            if not get_ini_res["indicator"]:
                logging.warning(f'Load INI File Error: {get_ini_res["message"]}')
                return prev_data
            ini_configs = get_ini_res["content"][MongoInitialware.MONGO]
            for section, configs in ini_configs.items():
                alias = section
                db_obj = self.mongo_handler_class(
                    alias=alias,
                    host=configs["host"],
                    port=int(configs["port"]),
                    username=configs["username"],
                    password=configs["password"],
                    database=configs["database"],
                    schemas=self.mongo_models[alias]
                )
                set_db(MongoInitialware.MONGO, section, db_obj)
        return prev_data

class PostgresInitialware(_DatabaseInitialware):
    """
    The PostgresInitialware class is used to initialize the postgres database.

    Args:
        postgres_handler_class (class): The postgres handler class.
        postgres_models (dict): {section: models}
    """
    def __init__(self, postgres_handler_class, postgres_models={}):
        self.postgres_handler_class = postgres_handler_class
        self.postgres_models = postgres_models
    
    def process(self, prev_data):
        """
        Initialize the postgres database.
        """
        if db_use_inspector():
            get_ini_res = self.get_ini_configs()
            if not get_ini_res["indicator"]:
                logging.warning(f'Load INI File Error: {get_ini_res["message"]}')
                return prev_data
            ini_configs = get_ini_res["content"][PostgresInitialware.POSTGRES]
            for section, configs in ini_configs.items():
                db_obj = self.postgres_handler_class(
                    host=configs["host"],
                    port=configs["port"],
                    username=configs["username"],
                    password=configs["password"],
                    database=configs["database"],
                )
                if section in self.postgres_models:
                    init_models = self.postgres_models[section]
                    keys_list = list(init_models.keys())
                    for index in range(len(keys_list)):
                        key = keys_list[index]
                        model = init_models[key]
                        if model.meta["initialize"]:
                            if model.meta["init_type"] == model.INIT_TYPE["SCHEMA"] and model.meta["authorization"] == None:
                                model.meta["authorization"] = configs["username"]
                            model.create_sql()
                        if index == len(keys_list) - 1:
                            model.execute_sql(db_obj)
                            model.clear_sql()
                set_db(PostgresInitialware.POSTGRES, section, db_obj)
        return prev_data

class JsonifyInitialware(Initialware):
    """
    The JsonifyInitialware class is used to define the method of jsonify.
    Some classes may not be able to use jsonify directly, so this class is used to define jsonify.
    
    Args:
        customized_jsonify (function): The customized jsonify function. The function must have two arguments: obj and default_jsonify.
            \- obj: The object to jsonify.
            \- default_jsonify: If the obj is not the type that the customized jsonify function can handle, you should use this function to jsonify the obj.
    
    Example:
        def customized_jsonify(obj, default_jsonify):
            if isinstance(obj, YourClass):
                return obj.to_json()
            return default_jsonify(obj)
    """
    def __init__(self, customized_jsonify=None):
        self.customized_jsonify = customized_jsonify
    
    def process(self, prev_data: dict):
        class CustomJSONEncoder(DefaultJSONProvider):
            @staticmethod
            def default_action(obj):
                if self.customized_jsonify:
                    return self.customized_jsonify(obj, default_jsonify=_default)
                else:
                    return _default(obj)
            default = default_action
        prev_data["app"].json = CustomJSONEncoder(prev_data["app"])
        return prev_data

class ImportI18NInitialware(Initialware):
    """
    The ImportI18NInitialware class is used to import the i18n data. Thus, server can use the i18n config to translate the message.
    The structure of the i18n json file is as follows:
    \n {
        "key1": {
            "en": "value1",
            "zh": "value2"
        },
        "key2": {
            "en": "value3",
            "zh": "value4"
        }
    }

    Args:
        file_path (str): The file path of the i18n json file. The default value is f"{os.getcwd()}/configs/public/i18n.json".
    """
    def __init__(self, file_path=f"{os.getcwd()}/configs/public/i18n.json"):
        self.file_path = file_path

    def process(self, prev_data):
        try:
            with open(self.file_path, "r", encoding="utf-8") as f:
                i18n_data = json.load(f)
        except:
            i18n_data = {}
        set_configs("I18N", i18n_data)
        
        return prev_data

class RegisterApisInitialware(Initialware):
    """
    The RegisterApisInitialware class is used to register the APIs. Based on the server_api.yaml, importing the necessary APIs to the server.
    Therefore, server does not need to import all the APIs in apis folder, which can reduce the memory usage and improve the efficiency.

    The process of registering the APIs is the same as the function api_route_register.
    Thus, the url_prefix is f"/api/{blueprint.name}".

    Args:
        api_folder (str): The folder name of the APIs. The default value is "apis".
        module_suffix (str): The suffix of the module. The default value is "_api".
        blueprint_suffix (str): The suffix of the blueprint. The default value is "Api".
        disabled_apis (list): The list of the disabled APIs. The default value is [].
    """
    def __init__(self, api_folder="apis", module_suffix="_api", blueprint_suffix="Api", disabled_apis=[]):
        self.api_folder = api_folder
        self.module_suffix = module_suffix
        self.blueprint_suffix = blueprint_suffix
        self.disabled_apis = disabled_apis
    
    def process(self, prev_data):
        SERVER_API = get_globals("API")
        API_CONFIGS = get_configs("API")
        API_ENABLE = API_CONFIGS.get("API_ENABLE", False)

        if API_ENABLE:
            api_list = SERVER_API.keys()
            for api_name in api_list:
                if api_name in self.disabled_apis:
                    continue
                api_module = importlib.import_module(f"{self.api_folder}.{api_name}{self.module_suffix}")
                api_blueprint = getattr(api_module, f"{api_name}{self.blueprint_suffix}")
                api_route_register(prev_data["app"], api_blueprint)
        return prev_data


import logging
from .global_tools import get_configs, get_globals

def global_use_inspector():
    """
    This function is used to check whether the GLOBALS is used.
    """
    return len(get_globals().keys()) != 0

def config_use_inspector():
    """
    This function is used to check whether the CONFIGS is used.
    """
    if len(get_configs().keys()) == 0:
        logging.debug("CONFIGS is not used.")

def api_use_inspector(request):
    """
    This function is used to check whether the server_api.yml is used and the request is from the API.
    """
    config_use_inspector()
    SERVER_CONFIGS_API = get_configs("API")
    return SERVER_CONFIGS_API.get("API_ENABLE", False) and request.path.startswith("/api") 

def cors_use_inspector():
    """
    This function is used to check whether the CORS is used.
    """
    config_use_inspector()
    CORS_CONFIGS = get_configs("CORS")
    return CORS_CONFIGS.get("CORS_ENABLE", False)

def db_use_inspector():
    """
    This function is used to check whether the DB is used.
    """
    config_use_inspector()
    DB_CONFIGS = get_configs("DB")
    return DB_CONFIGS.get("DB_ENABLE", False)

def jwt_use_inspector():
    config_use_inspector()
    JWT_CONFIGS = get_configs("JWT")
    if type(JWT_CONFIGS) == dict:
        return JWT_CONFIGS.get("JWT_ENABLE", False)
    else:
        return False